#!/home/moerman/cadbin/perl -w
# Maze generator
#
# This script is a maze (Dutch: doolhof) generator, which can
# generate arbitrary-sized mazes. Lot of options to tune things,
# however not all work well together yet (e.g. opc_c=2, opt_e>0).
#
# To understand the program, first see the simple_maze.pl maze generator.
# Same algorithm, but with all options stripped so the bare
# principles are visible
#
  my $help_text = <<HELP;		# inline help text placed here
# $0 version 1.0, C.M. Moerman (c) April 2001
#
# Syntax: perl -w doolhof.pl [ -options ]
# Where:
#   -?      displays this help message (and quits)
#   -a1     give some air (don't fill all holes)
#   -c1     color output instead of black and white (-c1, -c2)
#   -e<n>   easy: punch extra holes (approx <n>), default 0
#   -f<f>   output file name, default -pdoolhof.png
#   -g<n>   graphical cell size, default 3
#   -l<n>   long lines (determines visual appearance) 1..10, def 0
#   -p1     print on stdout instead of .png graphic file
#   -r<n>   repeat number of gaps, default 1 (only 0/1 implemented)
#   -s<n>   size of internal gap, default 8
#   -t1     don't add (copyright) title
#   -x<n>   graphical nr of cells in horizontal direction, default 20
#   -y<n>   graphical nr of cells in vertical direction, default 10
#   -z<n>   random seed value (default random)
# Separate options with spaces, e.g. -x100 -y40 -l1 -a1 -fmaze.png
#
HELP
#
# How does it work: if you have a room with external walls only and two
# doors, there is one way from one door to the next. Note that this does
# not change if you add one wall starting from an existing wall, but not
# connecting to anything else. This way, you can keep on adding walls,
# with only one side connected, and still keep one (and only one) route
# from the input to the output. The route just will get more complex, and
# more and more false paths are added.
# 
# This is also how this program created mazes. We start with an empty
# space (by init_maze), add the external walls (init_borders), and just
# start on adding walls. This is first done at random: find a free spot
# and from there find a wall (find_free_pole), and start creating walls at
# that point (draw_segments). As we started at a free spot, there is
# always a direction to add walls.
# 
# In the last stage, we systematically scan for free points and draw walls
# (check_all_poles), just in case we have missed any areas with the random
# filling.
# 
# This is all done on the representation using the visual metaphore of
# having a grid with poles, and tying ropes (segments) between the poles
# as walls. Remember, when we place N poles, we have only N-1 ropes needed
# between the poles.
#
# Kees Moerman, April 2, 2001
#
# Change list:
#	C.M.Moerman, April 5, 2001: several options added
#   C.M.Moerman, April 9, 2001: 'l' option added, general clean-up
#
# To do:
#

use strict;                             # don't forget to prefix vars with my

use GD;                                 # graphics library



# --------------- command line handling and option setting -----------

my $opt_a = 0;                          # default -a0   1 = 'airy' look (don't fill all holes)
my $opt_c = 0;                          # default -c0   1, 2 colouring scheme
my $opt_e = 0;                          # default -e0   1 = 'easy' level (puch extra holes)
my $opt_f = "doolhof.png";              # default file name
my $opt_g = 3;                          # default -g3   graphical cell size
my $opt_l = 0;                          # default -l0   long line size
my $opt_p = 0;                          # default -p0   1 = print instead of png file
my $opt_r = 1;                          # default -r1   number of central gaps
my $opt_s = 8;                          # default -s8   size of central gap
my $opt_t = 0;                          # default -t0   1 = don't put title below png
my $opt_x = 20;                         # default -x10  number of cells
my $opt_y = 10;                         # default -y10  number of cells
my $opt_z = 0;                          # default no random seed supplied

$help_text =~ s/^# ?//gm;               # clean up comment characters in help text

if (!defined $ARGV[0] || $ARGV[0] eq "-?" || $ARGV[0] eq "-h"  )
{   print $help_text;
    exit 1;
}

while ($ARGV[0] =~ /^-(.)(.+)/)
{
    $opt_a = $2  if ($1 eq "a");        # -a<n>: airy look?
    $opt_c = $2  if ($1 eq "c");        # -c<n>: color?
    $opt_e = $2  if ($1 eq "e");        # -e<n>: easy level
    $opt_f = $2  if ($1 eq "f");        # -f<f>: filename
    $opt_g = $2  if ($1 eq "g");        # -g<n>: graphic cell size
    $opt_l = $2  if ($1 eq "l");        # -l<n>: long lines
    $opt_p = $2  if ($1 eq "p");        # -p<n>: print?
    $opt_r = $2  if ($1 eq "r");        # -r<n>: repeat gaps
    $opt_s = $2  if ($1 eq "s");        # -s<n>: gap size
    $opt_t = $2  if ($1 eq "t");        # -p<n>: title?
    $opt_y = $2  if ($1 eq "y");        # -y<n>: number of cells in y direction
    $opt_x = $2  if ($1 eq "x");        # -x<n>: number of cells in x direction
    $opt_z = $2  if ($1 eq "z");        # -z<n>: random seed
    shift;
    last if !defined $ARGV[0];
}

# ----------------------------- data structures -------------------

# representation of the maze is done in two arrays; one for the
# horizontal lines and one for the vertical ones. Look at the below
# figure for the orientation and numbering of the elements. See the
# maze as a number of poles in a grid, with lines drawn between them.
#
#      0           1           2
#    
# 0    X--h[0,0]---X--h[1,0]---X    h --> actually array @horlines
#      |           |           |
#   v[0,0]      v[1,0]      v[2,0]
#      |           |           |
# 1    X--h[0,1]---X--h[1,1]---X
#      |           |           |
#   v[0,1]      v[1,1]      v[2,1]
#      |           |           |
# 2    X--h[0,2]---X--h[1,2]---X


sub max { ($_[0]>$_[1])?$_[0]:$_[1]; }
sub min { ($_[0]<$_[1])?$_[0]:$_[1]; }

my $xmax = min(max($opt_x & 0xFFFE, 10), 1000);   # number of lines (poles: +1)
my $ymax = min(max($opt_y & 0xFFFE, 10), 1000);   # force even
  $opt_g = min(max($opt_g, 2), 10);
my $size = min(max($opt_s & 0xFFFE, 4), min($xmax/2, $ymax/2));

my @horlines;                           # note: X poles need X-1 lines 
my @verlines;                           # see above diagram for interpretation

# hide internal representation with access functions

sub set_hor { my ($x, $y, $z) = (@_); $horlines[$x][$y] = $z; } # $z = color
sub set_ver { my ($x, $y, $z) = (@_); $verlines[$x][$y] = $z; }
sub get_hor { my ($x, $y) = (@_);     return $horlines[$x][$y]; }
sub get_ver { my ($x, $y) = (@_);     return $verlines[$x][$y]; }

my ($gx, $gy, $color, $pref_dir);       # global location used by some functions

# ------------------------------ initialisation ----------------------

sub init_maze                           # initialise edge arrays with zeros
{   my ($x, $y);                        # ( ) --> 
    for $x (0..$xmax)
    {   for $y (0..$ymax)
        {   set_hor($x, $y, 0);
            set_ver($x, $y, 0);
        }
    }
}

sub init_borders                        # initialise boundaries at 1
{   my ($x, $y);                        # ( ) --> 
    for $x (0..$xmax-1)                 # X poles -> X-1 lines
    {   
        set_hor($x, 0, 1);
        set_hor($x, $ymax, 1);
    }
    for $y (0..$ymax-1)
    {
        set_ver(0, $y, 1);
        set_ver($xmax, $y, 1);
    }
}

# --------------------------- basic subroutines ----------------------

sub rnd {   int (rand( $_[0] )); }      # random 0..N-1: ( N ) --> ( int )

sub is_pole_free                        # check if pole is standing free
{   my ($x, $y) = (@_);                 # ( x, y ) --> ( flag )
                                        # check boundaries first
    return 0 if (($x<0) || ($y<0) || ($x>=$xmax) || ($y>=$ymax));
    return 0 if get_ver($x, $y-1);      # line on top
    return 0 if get_hor($x-1, $y);      # line to the left
    return 0 if get_ver($x, $y);        # line at bottom
    return 0 if get_hor($x, $y);        # line to the right
    return 1;
}

sub draw_next_segment                   # add one more segment (if possible)        
{   my $temp = rnd(4);                  # on current line, random direction
                                        # operate on globals $gx, $gy, $color
    if($pref_dir>=0)                    # OK, did we find a preference direction?
    {                                   # in that case, give slight advantage
        $temp = rnd(4+$opt_l);          # to the previously registered direction
        $temp = $pref_dir & 3 if ($temp > 3);   # strength determined by $opt_l
    }
    
    for (0..1)                          # might need to re-iterate once
    {
        if (($temp & 3) == 0)           # left?
        {   if(is_pole_free($gx-1, $gy))
            { set_hor($gx-1, $gy, $color); $gx--; $pref_dir = $temp if $pref_dir<0; return 1}
            $temp++;                    # failed, try next direction
        }
            
        if (($temp & 3) == 1)           # down
        {   if(is_pole_free($gx, $gy+1))
            { set_ver($gx, $gy, $color); $pref_dir = $temp if $pref_dir<0; $gy++; return 1 }
            $temp++;
        }
            
        if (($temp & 3) == 2)           # right
        {   if(is_pole_free($gx+1, $gy))
            { set_hor($gx, $gy, $color); $pref_dir = $temp if $pref_dir<0; $gx++; return 1 }
            $temp++;
        }
            
        if (($temp & 3) == 3)           # up
        {   if(is_pole_free($gx, $gy-1))
            { set_ver($gx, $gy-1, $color); $pref_dir = $temp if $pref_dir<0; $gy--; return 1 }
            $temp++;
        }
    }
    return 0;
}

sub draw_segments                       # generate one curly line  
{   ($gx, $gy) = (@_);                  # ( x, y ) -->
                                        # set and work on globals $gx, $gy
    $color = 7 if $opt_c == 2;          # just to detect errors, should not show up
    $color = get_ver($gx, $gy  ) if (get_ver($gx, $gy  ) && $opt_c == 2);
    $color = get_ver($gx, $gy-1) if (get_ver($gx, $gy-1) && $opt_c == 2);
    $color = get_hor($gx,   $gy) if (get_hor($gx,   $gy) && $opt_c == 2);
    $color = get_hor($gx-1, $gy) if (get_hor($gx-1, $gy) && $opt_c == 2);

    $pref_dir = -1;                     # not yet a preference direction
    while( rnd($xmax)!=0)               # limit max length
    {
        last if !draw_next_segment();   # as long as we can add segments
    }
}

sub move_near_free                      # find first occupied pole to the left/right/...
{                                       # input in globals: free standing pole
    my $rand = rnd(4);                  # look in random direction

    if( $rand == 0)                     # find first free pole to left
    {   $gx-- while( is_pole_free($gx-1, $gy));
        $gx--;                          # so this one connected to walls
    }
    elsif ($rand == 1)                  # find first free pole to top
    {   $gy++ while( is_pole_free($gx, $gy+1));
        $gy++;                          # so this one connected to walls
    }
    elsif( $rand == 2)                  # first free pole to bottom
    {   $gy-- while( is_pole_free($gx, $gy-1));
        $gy--;                          # so this one connected to walls
    }
    else                                # first free pole to right
    {   $gx++ while( is_pole_free($gx+1, $gy));
        $gx++;                          # so this one connected to walls
    }
}

sub find_free_pole                      # find random free-standing pole close to wall
{   for (0..10)                         # return in globals gx, gy
    {
        $gx = rnd($xmax-1);             # (try max 10 times to find is_pole_free spot)
        $gy = rnd($ymax-1);
        if( is_pole_free($gx, $gy))     # random free pole
        {
            move_near_free();           # so find close wall (any direction is OK)
            return 1;
        }
    }
    return 0;                           # not found
}

sub check_all_poles                     # check/fill if still empty pole
{   my ($x, $y);                        # carefull: don' use $gx as global and as loop
    for $x (1..$xmax-1)                 # using exhaustive check on globals gx, gy
    {
        for $y (1..$ymax-1)
        {   $gx = $x; $gy = $y;
            if(is_pole_free($gx, $gy))  # OK, found freestanding pole
            {
                move_near_free();       # get close wall and try to connect
                draw_segments($gx, $gy);    # may go in wrong direction, though!)
            }
        }
    }
}

# ------------------------- now for the fancy extras -------------------------


sub hline                               # straigth horizontal line
{                                       # ( x, y, len ) -->
    my ($x, $y, $len) = (@_);
    for (0.. $len)
    {
        set_hor($x++, $y, $color);
    }
}

sub vline                               # straigth horizontal line
{                                       # ( x, y, len ) -->
    my ($x, $y, $len) = (@_);
    for (0.. $len)
    {
        set_ver($x, $y++, $color);
    }
}

sub hcurve                              # wiggly horizontal line
{                                       # ( x, y, len ) -->
    my ($x, $y, $len) = (@_);
    for (0.. $len/2)
    {
        if(rnd(2))                      # random wiggle totop or bottom
        {
            set_hor($x++, $y, $color);  #     +--+  +--+  +--+
            set_ver($x, $y-1, $color);  #   --+  +--+  +--+  +
            set_hor($x++, $y-1, $color);
            set_ver($x, $y-1, $color);
        }
        else
        {
            set_hor($x++, $y, $color);
            set_ver($x, $y, $color);
            set_hor($x++, $y+1, $color);
            set_ver($x, $y, $color);
        }
        
    }
}

sub vcurve                              # wiggly vertical line (x, y, len)
{                                       # ( x, y, len ) -->
    my ($x, $y, $len) = (@_);
    for (0.. $len/2)
    {
        if(rnd(2))                      # random wiggle to left or right
        {
            set_ver($x, $y++, $color);
            set_hor($x-1, $y, $color);
            set_ver($x-1, $y++, $color);
            set_hor($x-1, $y, $color);
        }
        else
        {
            set_ver($x, $y++, $color);
            set_hor($x, $y, $color);
            set_ver($x+1, $y++, $color);
            set_hor($x, $y, $color);
        }
    }
}


sub make_square                         # make inner square ($opt_r)
{   vline($xmax/2-$size/2, $ymax/2-$size/2, $size-1);
    vline($xmax/2+$size/2, $ymax/2-$size/2, $size-1);
    hline($xmax/2-$size/2, $ymax/2-$size/2, $size-1);
    hline($xmax/2-$size/2, $ymax/2+$size/2, $size-1);
    vcurve($xmax/2, 0, $ymax/2-$size/2);
    vcurve($xmax/2, $ymax/2+$size/2, $ymax/2-$size/2);
}

sub clean_square                        # make inner square doors, clean up
{   my ($x, $y);                        # ( ) -->
    set_ver($xmax/2-$size/2, $ymax/2, 0);   # open doors
    set_ver($xmax/2+$size/2, $ymax/2, 0);
    $color = 0;
    for $x (1..$size-1)                     # clean inner square
    {   vline ($xmax/2-$size/2+$x, $ymax/2-$size/2, $size-1);
    }
    for $y (1..$size-1)
    {   hline ($xmax/2-$size/2, $ymax/2-$size/2+$y, $size-1);
    }
}

sub easy                                # easy level opt_e (clear extra walls)
{   my ($x, $y);                        # ( ) -->
    for (0..$opt_e)
    {
        $x = rnd($xmax-2)+1;            # punch random holes
        $y = rnd($ymax-2)+1;            # (but not in borders, check value!)
        set_ver($x, $y, 0) if get_ver($x, $y) > 2; 
        set_hor($x, $y, 0) if get_hor($x, $y) > 2;
    }
}

sub color_walls                         # used for opt_c=2: show solution
{   my ($x, $y);                        # ( ) -->
    for $x (0..$xmax)
    {   for $y (0..$ymax/2)
        {   set_hor($x, $y, 3) if get_hor($x, $y);
            set_ver($x, $y, 3) if get_ver($x, $y);
        }
    }
    for $x (0..$xmax)
    {   for $y ($ymax/2..$ymax)
        {   set_hor($x, $y, 4) if get_hor($x, $y);
            set_ver($x, $y, 4) if get_ver($x, $y);
        }
    }
}


# ------- and the actual maze generation and printing ---------------------

sub doolhof                             # generate complete doolhof
{                                       # ( ) -->
    $color = 1;                         # colouring for debug purposes ($opt_c=1)
    init_maze();                        # or for showing solution (opt_c = 2)
    init_borders();                     # so external borders set up

    $color = 2;
    make_square() if $opt_r;            # internal borders around gap
    color_walls() if $opt_c == 2;       # and give colouring for solution

    $color = 3;
    for (0..sqrt($xmax+$ymax))          # some initial walls starting at border
    {                                   # is this actually needed?
        draw_segments( rnd($xmax), 0);
        draw_segments( rnd($xmax), $ymax);
        draw_segments( 0, rnd($ymax));
        draw_segments( $xmax, rnd($ymax));
    }

    $color = 4;
    while ( find_free_pole())           # start random filling
    {
        draw_segments( $gx, $gy);
    }
    
    $color = 5;
    check_all_poles;                    # systematic search
    
    $color = 6;
    easy() if $opt_e;                   # make punch some extra holes
    check_all_poles if !$opt_a;         # systematic search once more
    easy() if $opt_e;                   # make punch some extra holes
    check_all_poles if !$opt_a;         # systematic search once more

    set_ver(0, $ymax/2, 0);             # clear external wall doors
    set_ver($xmax, $ymax/2, 0);
    clean_square() if $opt_r;           # clear internal gap
}

sub visualise                           # make ASCII representation
{   my ($x, $y, $temp);                 # ( ) --> ( string )

    $temp = "";                         # poles are represented by '+'
    for $y (0..$ymax-1)                 # vertical waals by '-'
    {                                   # horizontals by '|'
        for $x (0..$xmax-1)             # and the suqares between poles by ' '
        {   
            $temp .= "+" . (get_hor($x, $y)?"-":" ");
        }
        $temp .= "+\n";
        for $x (0..$xmax)
        {   
            $temp .= (get_ver($x, $y)?"|":" ") . " ";
        }
        $temp .= "\n";
    }
    for $x (0..$xmax-1)
    {   
        $temp .= "+" . (get_hor($x, $ymax)?"-":" ");
    }
    return "$temp+\n";
}

sub make_png                            # make graphical representation
{                                       # ( ) -->
    my $im = new GD::Image($xmax*$opt_g+1,$ymax*$opt_g+1+($opt_t?15:0));

    my $white   = $im->colorAllocate(255,255,255);  # allocate some colors
    my $black   = $im->colorAllocate(0,0,0);       
    my $red     = $im->colorAllocate(255,0,0);      
    my $green   = $im->colorAllocate(0,255,0);
    my $blue    = $im->colorAllocate(0,0,255);
    my $cyan    = $im->colorAllocate(0,255,255);
    my $magenta = $im->colorAllocate(255,0,255);
    my $yellow  = $im->colorAllocate(255,255,0);    # debugging only

    my @lookup = ($white, $black, $red, $green, $blue, $cyan, $magenta, $yellow);
    @lookup = (1, 1, 1, 1, 1, 1, 1, 1, 1, 1) if !$opt_c;    # override to B/W

    $im->fill(1,1,$white);              # And fill bitmap with white
    
    my ($x, $y);

    for $y (0..$ymax-1)                 # just like in visualise, but now graphical
    {                                   # taking into acount $opt_g for cell size
        for $x (0..$xmax-1)
        {   
            $im->line($x*$opt_g, $y*$opt_g, ($x+1)*$opt_g, $y*$opt_g, $lookup[get_hor($x, $y)])
                if get_hor($x, $y);
        }
        for $x (0..$xmax)
        {   
            $im->line($x*$opt_g, $y*$opt_g, $x*$opt_g, ($y+1)*$opt_g, $lookup[get_ver($x, $y)])
                if get_ver($x, $y);
        }
    }
    for $x (0..$xmax-1)
    {   
        $im->line($x*$opt_g, $ymax*$opt_g, ($x+1)*$opt_g, $ymax*$opt_g, $lookup[get_hor($x, $ymax)])
            if get_hor($x, $ymax);
    }

    $im->string(($xmax*$opt_g>=360)?gdSmallFont:gdTinyFont, 2, $ymax*$opt_g+2,
        ($xmax*$opt_g>=180)?"Kees Moerman (c) 2001 (${xmax}x$ymax), $opt_z":"KM ${xmax}x$ymax $opt_z" ,
         $black)
        if !$opt_t;

    my $png_data = $im->png;            # OK, made complete bitmap, not write it out
    open (DISPLAY,">$opt_f") || die "Can't write data to $opt_f: $!";
    binmode DISPLAY;
    print DISPLAY $png_data;
    close DISPLAY;
}


# ------------------------------ 'main' program ---------------------

srand($opt_z) if $opt_z;                # initialise random generator

doolhof();                              # generate maze
print visualise() if $opt_p;
make_png() if !$opt_p;


#end of perl script

